(*pp deriving *)
(** Core types *)

module type FableType =
sig
  type 'a expression'
  type 'a datatype'
  type untyped_data
  type typed_data

  exception Bad_fable_type of string

  type kind = MKind | UKind 
      deriving (Eq, Show, Pickle, Typeable, Shelve)

  type 'a lbasis = [ `Nothing | `Something of 'a]
      deriving (Eq, Show, Pickle, Typeable, Shelve)
      
  type 'a triopt = 
      [ `Nothing  
      | `Inferred of 'a lbasis Unionfind.point 
      | `Required of 'a]
	deriving (Eq, Show, Pickle, Typeable, Shelve)
	 
  type 'a fable_qualifier = 
      [ `Empty
      | `KindVar of 'a kind_var_basis Unionfind.point
      | `Kind of kind 
      | `Name of string * 'a fable_qualifier
      | `Labeling of 'a expression' * 'a fable_qualifier
      | `Label of 'a expression' triopt * 'a fable_qualifier
      | `Phantom of string list * 'a fable_qualifier ]
  and 'a kind_var_basis = 
      [ `FlexKind of int * kind (*bound*)
      | `FixedKind of 'a fable_qualifier ]
(*       deriving (Eq, Show, Pickle, Typeable, Shelve) *)
	
  type untyped_expression = untyped_data expression'
  and unchecked_fable_type = untyped_data fable_qualifier 
	
  type typed_expression = typed_data expression'
  and fable_type = typed_data fable_qualifier
     
  val kind_leq : (kind*kind) -> bool
  val kind_qual : 'a fable_qualifier -> kind
  val kind_join : (kind*kind) -> kind
  val well_formed_fq : 'a fable_qualifier -> bool
  val instantiate : 'a fable_qualifier -> 'a fable_qualifier ->'a fable_qualifier
  val combine : 'a fable_qualifier -> 'a fable_qualifier ->'a fable_qualifier
  val def_annot : unit -> 'a fable_qualifier
  val name_annot : string -> 'a fable_qualifier
  val labeling_annot : 'a expression' -> 'a fable_qualifier 
  val label_annot : 'a expression' triopt -> 'a fable_qualifier 
  val phantom_annot : string list -> 'a fable_qualifier 
  val string_as_kind : string -> kind
  val kind_as_annot : kind -> 'a fable_qualifier

end

type 'a stringmap = 'a Utility.StringMap.t
type 'a field_env = 'a stringmap deriving (Eq, Pickle, Typeable, Show, Shelve)

module type T = 
sig
  type typed_data
  type untyped_data
  type kind
  type 'a fq
    
  (* type var sets *)
  module TypeVarSet : Utility.INTSET

  (* points *)
  type 'a point = 'a Unionfind.point 

  module Pickle_point (A : Pickle.Pickle) : Pickle.Pickle with type a = A.a Unionfind.point

  type primitive = [ `Bool | `Int | `Char | `Float | `XmlItem | `DB
  | `Abstract | `NativeString]
      deriving (Eq, Typeable, Show, Pickle, Shelve)

      
  type 'a datatype' = {ltype: 'a ldatatype'; ftype : 'a fq}
  and meta_type_var_basis = Flexible of int
			    | Rigid of int
			    | UncheckedRecursive of (int * untyped_data datatype')
			    | Recursive of (int * typed_data datatype')
			    | Body of typed_data datatype'
  and meta_row_var_basis  = ClosedRow
			    | FlexibleRow of int
			    | RigidRow of int
			    | RecursiveRow of (int * typed_data row')
			    | BodyRow of typed_data row'
  and 'a ldatatype' =
      [ `Not_typed
      | `Primitive of primitive
      | `Function of ('a datatype' * 'a datatype' * 'a datatype')
      | `Record of 'a row'
      | `Variant of 'a row'
      | `Table of 'a datatype' * 'a datatype'
      | `Application of (string * ('a datatype') list)
      | `MetaTypeVar of meta_type_var]
  and 'a field_spec' = [ `Present of 'a datatype' | `Absent ]
  and 'a field_spec_map' = 'a field_spec' field_env
  and 'a row' = 'a field_spec_map' * row_var
  and row_var = meta_row_var_basis point
  and meta_type_var = meta_type_var_basis point

  (* Abbreviations *)
  type datatype = typed_data datatype'
  and  row = typed_data row'
  and  ldatatype = typed_data ldatatype'
  and  field_spec = typed_data field_spec'
  and  field_spec_map = typed_data field_spec_map'

  type u_datatype = untyped_data datatype'
  and  u_ldatatype = untyped_data ldatatype'
  and  u_field_spec = untyped_data field_spec'
  and  u_field_spec_map = untyped_data field_spec_map'
  and  u_row = untyped_data row'
      
  val concrete_type : datatype -> datatype

  type type_variable = [`TypeVar of int | `RigidTypeVar of int | `RowVar of int]
      deriving (Typeable, Show, Pickle)
  type quantifier = type_variable * kind
      deriving (Typeable, Show, Pickle)
      
  type 'a assumption' = (quantifier list * 'a datatype')
  type u_assumption = untyped_data assumption'
  type assumption =  (quantifier list * datatype)
      (*       deriving (Show) *)


  val def_fable_annot : 'a ldatatype' -> 'a datatype'
  val fable_annot : 'a ldatatype' -> 'a fq -> 'a datatype'
  val add_fable_annot : 'a datatype' -> 'a fq  -> 'a datatype'
    
  (* useful types *)
  val unit_type : unit -> 'a datatype'
  val bool_type : unit -> 'a datatype'
  val string_type : unit -> 'a datatype'
  val xml_type : unit -> 'a datatype'
  val native_string_type : unit -> 'a datatype'

  (* get type variables *)
  val free_type_vars : datatype -> TypeVarSet.t
  val free_row_type_vars : row -> TypeVarSet.t

  val free_bound_type_vars : datatype -> TypeVarSet.t
  val free_bound_row_type_vars : row -> TypeVarSet.t

  (* used to freshen mailboxes in typename aliases *)
  val freshen_mailboxes : datatype -> datatype

  (** Fresh type variables *)
  val fresh_raw_variable : unit -> int

  val bump_variable_counter : int -> unit

  (* type variable construction *)
  val make_type_variable : int -> datatype
  val make_rigid_type_variable : int -> datatype
  val make_row_variable : int -> row_var
    
  (* fresh type variable generation *)
  val fresh_type_variable : unit -> datatype
  val fresh_rigid_type_variable : unit -> datatype

  val fresh_row_variable : unit -> row_var
  val fresh_rigid_row_variable : unit -> row_var

  (** rows *)
  (* empty row constructors *)
  val make_empty_closed_row : unit -> 'a row'
  val make_empty_open_row : unit -> 'a row'

  (* singleton row constructors *)
  val make_singleton_closed_row : (string * 'a field_spec') -> 'a row'
  val make_singleton_open_row : (string * 'a field_spec') -> 'a row'

  (* row predicates *)
  val is_closed_row : 'a row' -> bool
  val is_absent_from_row : string -> 'a row' -> bool

  (* row_var retrieval *)
  val get_row_var : row -> int option

  (* row update *)
  val row_with : (string * 'a field_spec') -> 'a row' -> 'a row'

  (* constants *)
  val empty_field_env : unit -> 'a field_spec_map'
  val closed_row_var : unit -> row_var

  val make_tuple_type : datatype list -> datatype
  val make_tuple_utype : u_datatype list -> u_datatype
    
  val field_env_union : ('a field_spec_map' * 'a field_spec_map') -> 'a field_spec_map'

  val contains_present_fields : 'a field_spec_map' -> bool

  val is_canonical_row_var : row_var -> bool
  val is_rigid_row : row -> bool

  val is_rigid_row_with_var : int -> row -> bool

  val is_flattened_row : row -> bool
  val is_empty_row : row -> bool

  (** Convert a row to the form (field_env, row_var)
      where row_var is of the form:
      [ `Closed
      | `Flexible var
      | `Rigid var
      | `Recursive
      ]
  *)
  val flatten_row : row -> row

  (**
     As flatten_row except if the flattened row_var is of the form:

     `Recursive (var, rec_row)

     then it is unwrapped. This ensures that all the fields are exposed
     in field_env.

     Also returns the outermost `Recursive that was unwrapped if it exists,
     or None otherwise.
  *)
  val unwrap_row : row -> (row * row_var option)
  val unwrap_row' : 'a row' -> (row * row_var option)

  type type_alias_set = Utility.StringSet.t

  val type_aliases : datatype -> type_alias_set
  val row_type_aliases : row -> type_alias_set
    
  type dictionary = 
      { s_of_fq : 'a.'a fq -> string -> string;
	    s_of_d : 'a. dictionary -> TypeVarSet.t -> string Utility.IntMap.t -> 'a datatype' -> string;
        s_of_r : 'a. dictionary -> string -> TypeVarSet.t -> string Utility.IntMap.t -> 'a row' -> string;
        is_closed_row : 'a. dictionary -> TypeVarSet.t -> 'a row' -> bool;}
  val dictionary : dictionary
    
  (* pretty printing *)
  val string_of_datatype : dictionary -> 'a datatype' -> string
  val string_of_datatype_raw : dictionary -> 'a datatype' -> string
  val string_of_row : dictionary -> 'a row' -> string
  val string_of_row_var : row_var -> string
  val string_of_assumption : dictionary -> 'a assumption' -> string
    
  val make_fresh_envs : datatype -> datatype Utility.IntMap.t * row_var Utility.IntMap.t
  val make_rigid_envs : datatype -> datatype Utility.IntMap.t * row_var Utility.IntMap.t
  val make_wobbly_envs : datatype -> datatype Utility.IntMap.t * row_var Utility.IntMap.t
    
  val functor_type : ('a fq -> 'b fq) -> 'a datatype' -> 'b datatype'
  val functor_row : ('a fq -> 'b fq) -> 'a row' -> 'b row'
end
  
module Make (F : FableType) : (T  with type 'a fq = 'a F.fable_qualifier
				  and type typed_data = F.typed_data
				  and type untyped_data = F.untyped_data
				  and type kind = F.kind)
  
