(** Provides utility function to transform database queries.
	@version 2.0
	@author Gilles Dubochet *)

open Syntax
open Expression

(** Types involved in backtracing variable origins *)
type projection_source = {
  field_name : string;
  field_var : string;
  etc_var : string;
  source_var : string;
}

(** A variable declaration stub. *)
type binding =
    [ `Calculated of string * Syntax.expression
        (** The variable is calculated (that is, NOT a simple record selection 
            on a variable) *)
    | `QCalculated of string * Query.expression
        (** The variable that has already be calculated into a query expression *)
    | `Selected of projection_source
        (** The variable comes from a record selection operation with {ol
            {li the extracted field,} {li the variable bound to the field,}
            {li the variable bound to the row,} {li the source variable.} } *)
    | `Table_loop of string * Expression.Query.query
        (** The variable is the looping variable for a loop on a
            table. Second field is the SQL name of the table. *)
    | `Unavailable of string ]

(** bindings
    A list of bindings. This list can be used to reconstruct a record
    selection tower. *)
type bindings = binding list

(** The required precondition values to calculate a variable.  *)
type origin =
    [ `Earlier of string * projection_source list
        (** The variable is based on a variable defined before the
            bindings start, but requires the given list of selections
            before it can be used. *)
    | `Table of Query.query
        (** The variable represents a whole row of the given table *)
    | `QCalculated of Query.expression
    | `Table_field of string * string
        (** The variable is a field of the table loop *)
    | `Unavailable
        (** The variable cannot be used in a query *)
    ]

exception ColumnNotInQuery of string

(** Convert a LIKE expression to a string. *)
val like_as_string : (string * Result.result) list -> Query.like_expr -> string

(** make_sql
    Converts an expression from the constant/variable sublanguage into
    an SQL expression (with respect to the given environment,
    `bindings') *)
val make_sql : bindings -> Syntax.expression 
  -> (Query.expression * projection_source list) option

(** condition_to_sql
    Converts a Links condition into an SQL condition; should only be
    applied to expressions that have boolean type
    
    @param expr The candidate syntax expression.
    @param bindings The bindings available where the expression occurs.
        A `Table_loop element must be present.
    @return Some(condition, projs) if it can transform the input into SQL.
        `condition' is the condition, now in Query.expression format, 
        and projs is a list of record selection descriptors. These
        projections need to be wrapped around the Query in order that 
        the free vars of the Query should be properly bound. Returns 
        None if it can't create such a result. *)
val condition_to_sql :  Syntax.expression ->  bindings 
  -> (Query.expression * projection_source list) option

(** Projects the query on the empty set.
    @param query The query to nullify.
    @return The nullified query. *)
val null_query : Query.query -> Query.query

(** Restricts the columns to be selected from a table query by
      specifying all columns to keep. Fails if at least one
      of the column to keep was not selected in the original
      query.

    @param projs A list of the columns to keep in the query.
    @param query The query to modify.
    @return The modified query. *)
val project : string list -> Query.query -> Query.query

(** select
    Adds conditions to a query. If the {! Sql_transform.selectable} method
     is called on every positive and negative condition provided,
     this function should not fail.
    @param positives Conditions that whould be satisfied by any element
      of the result.
    @param negatives Conditions that should not be satisfied by any 
      element of the result.
    @param query The original query.
    @return The query with the added conditions.
    @raise Failure A general programming error has occured. Passing 
      conditions that where not selectable with {! Sql_transform.selectable}
      might cause such exceptions. *)
val select : Query.expression list -> Query.query -> Query.query

(** select_by_origin
    Adds the origin selections before an expression.
    @param origin The origin to add to the expression.
    @param expr The expression to modify.
    @return The query with the added origin. *)
val select_by_origin :  projection_source list -> Syntax.expression -> Syntax.expression

(** trace_variable
    Finds the origin of a given variable according to a given binding.
    @param name The name of the variable to lookup.
    @param bindings The binding list to search.
    @return  *)
val trace_variable : string -> bindings -> origin

(** sep_assgmts
    Given an expression that consists of several assignments (lets or
    record_selections) wrapped around an inner expression, sep_assgmts
    returns a pair of the inner expression with a list that shows
    where each bound variable came from. Currently it only handles
    trivial projections of the form x.a where x is a variable term;
    all other assignments are considered "unavailable".

    @param expr The candidate syntax expression.
    @param bindings The bindings available where the expression occures.
    @return {i true} if the expression can be used, {i false} otherwise. *)
val sep_assgmts :  bindings -> Syntax.expression -> bindings * Syntax.expression

(** join
    Joins two queries into one, over the given condtions. If the
    {! Sql_transform.selectable} method is called on every positive and
    negative condition provided, this function should not fail.
    @param positives Conditions that whould be satisfied by any element of 
       the result.
    @param negatives Conditions that should not be satisfied by any element 
       of the result.
    @param left_query The first original query.
    @param left_query The second original query.
    @return The join between queries with the added conditions.
    @raise Failure A general programming error has occured. Passing 
        conditions that where not selectable with 
    {! Sql_transform.selectable} might cause such exceptions. *)
val join : Query.expression list -> Query.query * Query.query 
  -> (Query.column * string) list * Query.query

(** pos_and_neg
    A simple utility that converts (pos, neg) to "pos AND NOT neg"
    One wonders why the (pos, neg) rep'n is used in the first place. *)
val pos_and_neg : Query.expression list * Query.expression list -> Query.expression
  
val call_stored_procs : bool Settings.setting
